#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['po4', 'po4.cli', 'po4.loaders', 'po4.stepwise']

package_data = \
{'': ['*']}

install_requires = \
['appdirs',
 'autoprop',
 'autosnapgene',
 'Biopython',
 'configurator',
 'docopt',
 'inform>=1.21',
 'more_itertools>=8.0',
 'pandas',
 'setuptools',
 'stepwise',
 'stepwise_mol_bio',
 'tabulate',
 'toml',
 'tqdm',
 'voluptuous',
 'xlrd>=1.0']

extras_require = \
{'docs': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'tests': ['pytest', 'pytest-cov', 'coveralls']}

entry_points = \
{'console_scripts': ['po4 = po4.cli.main:main'],
 'po4.databases': ['excel = po4.loaders.excel'],
 'stepwise.protocols': ['protocols = po4.stepwise:Plugin']}

setup(name='po4',
      version='0.2.0',
      description='Database of DNA sequences.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/po4',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='~=3.8',
     )
